(* Type représentant un trie.
   Pour ce type et la fonction find, voir:

    https://gist.github.com/Tchou/c0c305621bc909e189db90821d78eb6f
*)

type 'a trie =
    Node of ('a option * (char * 'a trie) list)
  (** [empty] est le trie vide. *)
let empty = Node (None, [])

(** [is_empty t] est vrai si et seulement si [t] est vide. *)
let is_empty t =
  match t with
    Node (None, []) -> true
  | _ -> false

(** [find_trie key t] renvoie le (sous)-trie accessible
    dans [t] en suivant la clé [key].
    Lève l'exception [Not_found] si [key] n'est pas dans [t] *)
let find_trie key t =
  let rec find_node i t =
    if i = String.length key then t
    else
      match t with Node (_, l) -> find_list i l

  and find_list i l =
    let ci = key.[i] in
    match l with
      [] -> raise Not_found
    | (d, t) :: ll ->
      if ci > d then find_list i ll
      else if ci = d then find_node (i+1) t
      else raise Not_found
  in
  find_node 0 t

(** [find key t] renvoie la valeur accessible dans
    [t] en suivant la clé [key]. Lève l'exception [Not_found]
    si [key] n'est pas dans [t]. *)
let find key t =
  match find_trie key t with
    Node (Some v, _) -> v
  | Node (None, _) -> raise Not_found

(** Q1.1 On suit la même structure que find. *)
let add key v t =
  let rec insert_node i t =
    match t with
    | Node (o, l) ->
      if i = String.length key then Node (Some v, l)
      else Node (o, insert_list i l)
  and insert_list i l =
    let ci = key.[i] in
    match l with
    | [] -> [ (ci, insert_node (i + 1) empty) ]
    | (d, t) :: ll ->
      if ci > d then (d, t) :: insert_list i ll
      else if ci = d then (ci, insert_node (i + 1) t) :: ll
      else (ci, insert_node (i + 1) empty) :: l (* attention, ici l pour laisser (d,t)*)
  in
  insert_node 0 t
(* Q1.2 Test : *)
(* On essaye d'écrire un test un peu générique *)
let add_from_list l t =
  List.fold_left (fun acc (k, v) -> add k v acc) t l

let test2_2 () =
  (* Une list des valeurs de test *)
  let l = [ "a", 10; "abc", 11; "abd", 12 ] in
  (* Remplissage du dico à partir de la liste *)
  let dico =  add_from_list l empty in
  (* On parcours la liste et on s'assure que la valeur extraite
     du dico pour cette clé est la valeur initialement associée.
  *)
  List.iter (fun (k, v) -> assert (find k dico = v)) l

let () = test2_2 () (* ne doit pas provoquer d'erreur *)


(* Q1.3 Ici, pas besoin de faire des fonctions imbriquées car par
   besoin de paramètre auxiliaire à initialiser: *)
let rec union f t1 t2 =
  match (t1, t2) with
  | Node (None, l1), Node (None, l2) -> Node (None, union_list f l1 l2)

  | Node (Some v, l1), Node (None, l2) (* attention 2 cas en 1 *)
  | Node (None, l1), Node (Some v, l2) -> Node (Some v, union_list f l1 l2)

  | Node (Some v1, l1), Node (Some v2, l2) -> Node (Some (f (v1, v2)), union_list f l1 l2)
and union_list f l1 l2 =
  match (l1, l2) with
  | [], l
  | l, [] ->
    l
  | (c1, t1) :: ll1, (c2, t2) :: ll2 ->
    if c1 < c2 then (c1, t1) :: union_list f ll1 l2
    else if c1 > c2 then (c2, t2) :: union_list f l1 ll2
    else (c1, union f t1 t2) :: union_list f ll1 ll2

(* Q1.4 *)
let test2_4 () =
  let keys1 = [ "A", 1; "B", 1; "ACAC", 1; "BB", 1 ] in
  let keys2 = [ "", 1; "ABBA", 1; "B", 1; "BA", 1] in
  let t_add = add_from_list keys2 (add_from_list keys1 empty) in
  let t_union = union snd (add_from_list keys1 empty)
      (add_from_list keys2 empty)
  in
  assert (t_add = t_union)

let () =  test2_4 ()
(* Q1.5
   Si on réfléchit, il n'y a qu'un chemin possible pour une clé,
   quelle que soit la façon de l'insérer et donc, pour un ensemble
   de clé donné, un seul trie possible.

   Ce n'est pas le cas des tables de hachages ou des AVL, qui peuvent
   avoir des formes différentes en fonction de l'ordre d'insertion et du
   moment où ont été fait les redimentionnement ou ré-équilibrages.

*)

